/*
<:copyright-broadcom 
 
 Copyright (c) 2002 Broadcom Corporation 
 All Rights Reserved 
 No portions of this material may be reproduced in any form without the 
 written permission of: 
          Broadcom Corporation 
          16215 Alton Parkway 
          Irvine, California 92619 
 All information contained in this document is Broadcom Corporation 
 company private, proprietary, and trade secret. 
 
:>
*/
//**************************************************************************
// File Name  : bcmenet.c
//
// Description: This is Linux network driver for the BCM63xx Ethernet MACs
//               
//**************************************************************************


#define VERSION     "0.3"

#define VER_STR     "v" VERSION " " __DATE__ " " __TIME__

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/errno.h>
#include <linux/delay.h>

#include <linux/mii.h>
#include <linux/skbuff.h>
#include <linux/ethtool.h>

#include <board.h>
#include <bcm_common.h>
#include <bcm_map.h>
#include <bcm_intr.h>
#include "bcmenet.h"
#include "bcmmii.h"

#include <linux/stddef.h>

#ifdef USE_PROC
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/types.h>
#define PROC_ENTRY_NAME     "driver/eth%dinfo"
#endif

#include <asm/r4kcache.h>

#define POLLTIME_100MS      (HZ/10)
#define POLLCNT_1SEC        (HZ/POLLTIME_100MS)
#define POLLCNT_FOREVER     ((int) 0x80000000)

#define EMAC_TX_WATERMARK   32

#define VLAN_DISABLE         0
#define VLAN_ENABLE          1

#define ENET_POLL_DONE      0x80000000


typedef enum {
    TX_BD_RING1 = 0,
    TX_BD_RING2,
    RX_BD_RING,
    DMA_DESC_STATUS,
    BUF_POOL,
    EMAC_REGS,
    MISC_REGS,
}   DUMP_TYPE;

#define MAKE4(x)   ((x[3] & 0xFF) | ((x[2] & 0xFF) << 8) | ((x[1] & 0xFF) << 16) | ((x[0] & 0xFF) << 24))
#define MAKE2(x)   ((x[1] & 0xFF) | ((x[0] & 0xFF) << 8))

/* Prototypes */
static int bcm63xx_enet_open(struct net_device * dev);
static int bcm63xx_enet_close(struct net_device * dev);
static void bcm63xx_enet_timeout(struct net_device * dev);
static void bcm_set_multicast_list(struct net_device * dev);
static struct net_device_stats * bcm63xx_enet_query(struct net_device * dev);
static void bcm63xx_enet_poll_timer(unsigned long arg);
static int bcm63xx_enet_xmit(struct sk_buff * skb, struct net_device * dev);
static irqreturn_t bcm63xx_enet_isr(int irq, void *, struct pt_regs *regs);
static int bcm63xx_enet_poll(struct net_device * dev, int *budget);
static uint32 bcm63xx_rx(void *ptr, uint32 budget);
static void bcm63xx_enet_free_skb_or_data(void *param, void *obj, int free_flag);
static int bcm_set_mac_addr(struct net_device *dev, void *p);
static void write_mac_address(struct net_device *dev);
static void assign_rx_buffers(BcmEnet_devctrl *pDevCtrl, unsigned char *data);
static void perfectmatch_write(volatile EmacRegisters *emac, int reg, const uint8 * pAddr, bool bValid);
static void perfectmatch_update(BcmEnet_devctrl *pDevCtrl, const uint8 * pAddr, bool bValid);
static void clear_mib(volatile EmacRegisters *emac);
static int init_emac(BcmEnet_devctrl *pDevCtrl);
static void init_dma(BcmEnet_devctrl *pDevCtrl);
static int init_buffers(BcmEnet_devctrl *pDevCtrl);
static int bcm63xx_init_dev(BcmEnet_devctrl *pDevCtrl);
static int bcm63xx_uninit_dev(BcmEnet_devctrl *pDevCtrl);
#ifdef USE_PROC
static int bcm63xx_enet_dump(BcmEnet_devctrl *pDevCtrl, char *buf, int reqinfo);
static int dev_proc_engine(void *data,loff_t pos, char *buf);
static ssize_t eth_proc_read(struct file *file, char *buf, size_t count, loff_t *pos);
#endif
int __init bcm63xx_enet_probe(void);
static int bcmIsEnetUp(struct net_device *dev);
static void __exit bcmenet_module_cleanup(void);
static int __init bcmenet_module_init(void);
static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr);
static int bcm63xx_enet_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);

static struct net_device *eth_root_dev = NULL;

#ifdef DUMP_DATA
static void dumpHexData(unsigned char *head, int len);
/*
 * dumpHexData dump out the hex base binary data
 */
static void dumpHexData(unsigned char *head, int len)
{
    int i;
    unsigned char *curPtr = head;

    for (i = 0; i < len; ++i) {
        if (i % 16 == 0)
            printk("\n");       
        printk("0x%02X, ", *curPtr++);
    }
    printk("\n");
}
#endif

static inline void cache_wback_inv(unsigned long addr, unsigned long size)
{
    unsigned long end, a;
    unsigned long dc_lsize = current_cpu_data.dcache.linesz;

    a = addr & ~(dc_lsize - 1);
    end = (addr + size - 1) & ~(dc_lsize - 1);
    while (1) {
        flush_dcache_line(a);   /* Hit_Writeback_Inv_D */
        if (a == end)
            break;
        a += dc_lsize;
    }
}

/* --------------------------------------------------------------------------
    Name: bcm63xx_enet_open
 Purpose: Open and Initialize the EMAC on the chip
-------------------------------------------------------------------------- */
static int bcm63xx_enet_open(struct net_device * dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    ASSERT(pDevCtrl != NULL);

    TRACE(("%s: bcm63xx_enet_open\n", dev->name));

    pDevCtrl->emac->config |= EMAC_ENABLE;           

    pDevCtrl->dmaCtrl->controller_cfg |= DMA_MASTER_EN;

    pDevCtrl->rxDma->cfg |= DMA_ENABLE;

    BcmHalInterruptEnable(pDevCtrl->rxIrq);

    /* Start the network engine */
    netif_start_queue(dev);

    return 0;
}

/* --------------------------------------------------------------------------
    Name: bcm63xx_enet_close
 Purpose: Stop communicating with the outside world
    Note: Caused by 'ifconfig ethX down'
-------------------------------------------------------------------------- */
static int bcm63xx_enet_close(struct net_device * dev)
{
    int i;
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    ASSERT(pDevCtrl != NULL);

    TRACE(("%s: bcm63xx_enet_close\n", dev->name));

    netif_stop_queue(dev);

    BcmHalInterruptDisable(pDevCtrl->rxIrq);
   
    pDevCtrl->rxDma->cfg &= ~DMA_ENABLE;
    // wait the current packet to complete before turning off EMAC, otherwise memory corruption can occur.
    for(i=0; pDevCtrl->rxDma->cfg & DMA_ENABLE; i++) {
        // put the line below inside - it seems the signal can be lost and DMA never stops
        pDevCtrl->rxDma->cfg &= ~DMA_ENABLE;
        if (i >= 100) { break; }
        udelay(20);
    }

    pDevCtrl->emac->config |= EMAC_DISABLE;
    // wait the shut down process to complete. EMAC registers should't be modified before that.
    for(i=0; pDevCtrl->emac->config & EMAC_DISABLE; i++) {
        if (i >= 100) { break; }
        udelay(20);
    }

    return 0;
}

/* --------------------------------------------------------------------------
    Name: bcm63xx_enet_timeout
 Purpose: 
-------------------------------------------------------------------------- */
static void bcm63xx_enet_timeout(struct net_device * dev)
{
    ASSERT(dev != NULL);

    TRACE(("%s: bcm63xx_enet_timeout\n", dev->name));

    dev->trans_start = jiffies;

    netif_wake_queue(dev);
}

/* --------------------------------------------------------------------------
    Name: bcm_set_multicast_list
 Purpose: Set the multicast mode, ie. promiscuous or multicast
-------------------------------------------------------------------------- */
static void bcm_set_multicast_list(struct net_device * dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    TRACE(("%s: bcm_set_multicast_list: %08X\n", dev->name, dev->flags));

    /* Promiscous mode */
    if (dev->flags & IFF_PROMISC) {
        pDevCtrl->emac->rxControl |= EMAC_PROM;   
    } else {
        pDevCtrl->emac->rxControl &= ~EMAC_PROM;   
    }

    /* All Multicast packets */
    if (dev->flags & IFF_ALLMULTI) {
        pDevCtrl->emac->rxControl |= EMAC_ALL_MCAST;
    } else {
        pDevCtrl->emac->rxControl &= ~EMAC_ALL_MCAST;
    }
}

/* --------------------------------------------------------------------------
    Name: bcm63xx_enet_query
 Purpose: Return the current statistics. This may be called with the card 
          open or closed.
-------------------------------------------------------------------------- */
static struct net_device_stats *
bcm63xx_enet_query(struct net_device * dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    return &(pDevCtrl->stats);
}

/*
 * bcm63xx_enet_poll_timer: reclaim transmit frames which have been sent out
 */
static void bcm63xx_enet_poll_timer(unsigned long arg)
{
    BcmEnet_devctrl *pDevCtrl = (BcmEnet_devctrl *)arg;
    int linkState;

    pDevCtrl->linkstatus_polltimer++;
    /* Reclaim transmit Frames which have been sent out */
    bcm63xx_enet_xmit(NULL, pDevCtrl->dev);

    if ( pDevCtrl->linkstatus_polltimer >= POLLCNT_1SEC ) {
        linkState = bcmIsEnetUp(pDevCtrl->dev);
        if (linkState != pDevCtrl->linkState) {
            pDevCtrl->linkState = linkState;
            if ( linkState ) {
                pDevCtrl->MibInfo.ulIfLastChange = (jiffies * 100) / HZ;

                if (pDevCtrl->EnetInfo.ucPhyType == BP_ENET_EXTERNAL_SWITCH) {
                    pDevCtrl->MibInfo.ulIfSpeed = SPEED_100MBIT;
                } else {
                    unsigned long v = mii_read(pDevCtrl->dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_AUX_CTRL_STATUS);
                    if( (v & MII_AUX_CTRL_STATUS_SP100_INDICATOR) == MII_AUX_CTRL_STATUS_SP100_INDICATOR )
                        pDevCtrl->MibInfo.ulIfSpeed = SPEED_100MBIT;
                    else
                        pDevCtrl->MibInfo.ulIfSpeed = SPEED_10MBIT;
#if defined (CONFIG_BCM96338)
                    phy_tuning_mse(pDevCtrl->dev);
#endif
                }

                printk((KERN_CRIT "%s Link UP.\n"),pDevCtrl->dev->name);
            } else {
                pDevCtrl->MibInfo.ulIfLastChange = 0;
                pDevCtrl->MibInfo.ulIfSpeed = 0;
                printk((KERN_CRIT "%s Link DOWN.\n"),pDevCtrl->dev->name);
            }

            /* Wake up the user mode application that monitors link status. */
            kerSysWakeupMonitorTask();
        }
    }

    pDevCtrl->timer.expires = jiffies + POLLTIME_100MS;
    /* Joel */
//    add_timer(&pDevCtrl->timer);
}

/* --------------------------------------------------------------------------
    Name: bcm63xx_enet_xmit
 Purpose: Send ethernet traffic
-------------------------------------------------------------------------- */
static int bcm63xx_enet_xmit(struct sk_buff * skb, struct net_device * dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    ASSERT(pDevCtrl != NULL);
 
    /*
     * Obtain exclusive access to transmitter.  This is necessary because
     * we might have more than one stack transmitting at once.
     */
    local_bh_disable();

    /* Reclaim transmitted buffers */
    while (pDevCtrl->txFreeBds < NR_TX_BDS)  {
        if (pDevCtrl->txBds[pDevCtrl->txHeadIndex].status & DMA_OWN)
            break;

        dev_kfree_skb_any (pDevCtrl->txSkb[pDevCtrl->txHeadIndex]);
        if (++pDevCtrl->txHeadIndex == NR_TX_BDS)
            pDevCtrl->txHeadIndex = 0;

        pDevCtrl->txFreeBds++;
    }

    if (skb == NULL) {
        netif_wake_queue(dev);
        __local_bh_enable();
        return 0;
    }

    /* If we could not queue this packet, free it */
    if (pDevCtrl->txFreeBds == 0) {
        TRACE(("%s: bcm63xx_enet_xmit low on txFreeBds\n", dev->name));
        netif_stop_queue(dev);
        __local_bh_enable();
        return 1;
    }

#ifdef DUMP_DATA
    printk("bcm63xx_enet_xmit: len %d", skb->len);
    dumpHexData(skb->data, (skb->len < 32) ? skb->len : 32);
#endif

    cache_wback_inv((unsigned long)skb->data, skb->len);

    /* Decrement total BD count */
    pDevCtrl->txFreeBds--;

    if (pDevCtrl->txFreeBds == 0) {
        TRACE(("%s: bcm63xx_enet_xmit no transmit queue space -- stopping queues\n", dev->name));
        netif_stop_queue(dev);
    }

    /* Track sent SKB, so we can release them latter */
    pDevCtrl->txSkb[pDevCtrl->txTailIndex] = skb;

    pDevCtrl->txBds[pDevCtrl->txTailIndex].address = (uint32)VIRT_TO_PHY(skb->data);
    pDevCtrl->txBds[pDevCtrl->txTailIndex].length  = skb->len;

    /* advance BD pointer to next in the chain. */
    if (pDevCtrl->txTailIndex == (NR_TX_BDS - 1)) {
        pDevCtrl->txBds[pDevCtrl->txTailIndex].status = DMA_OWN | DMA_SOP | DMA_EOP | DMA_APPEND_CRC | DMA_WRAP;
        pDevCtrl->txTailIndex = 0;
    } else {
        pDevCtrl->txBds[pDevCtrl->txTailIndex].status = DMA_OWN | DMA_SOP | DMA_EOP | DMA_APPEND_CRC;
        pDevCtrl->txTailIndex++;
    }

    /* Enable DMA for this channel */
    pDevCtrl->txDma->cfg = DMA_ENABLE;

    /* update stats */
    pDevCtrl->stats.tx_bytes += ((skb->len < ETH_ZLEN) ? ETH_ZLEN : skb->len);
    pDevCtrl->stats.tx_bytes += ETH_CRC_LEN;
    pDevCtrl->stats.tx_packets++;

    dev->trans_start = jiffies;

    __local_bh_enable();

    return 0;
}

static int bcm63xx_enet_poll(struct net_device * dev, int * budget)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    uint32 work_to_do = min(dev->quota, *budget);
    uint32 work_done;
    uint32 ret_done;

    work_done = bcm63xx_rx(pDevCtrl, work_to_do);
    ret_done = work_done & ENET_POLL_DONE;
    work_done &= ~ENET_POLL_DONE;

    *budget -= work_done;
    dev->quota -= work_done;

    if (work_done < work_to_do && ret_done != ENET_POLL_DONE) {
        /* Did as much as could, but we are not done yet */
        return 1;
    }

    /* We are done */
    netif_rx_complete(dev);
    BcmHalInterruptEnable(pDevCtrl->rxIrq);
    return 0;
}

/*
 * bcm63xx_enet_phy_isr: Acknowledge interrupts and set flag for poll timer to
 *                       check link state.
 */
static irqreturn_t bcm63xx_enet_phy_isr(int irq, void * dev_id, struct pt_regs * regs)
{
    BcmEnet_devctrl *pDevCtrl = dev_id;

    mii_clearphyinterrupt(pDevCtrl->dev, pDevCtrl->EnetInfo.ucPhyAddress);
    pDevCtrl->linkstatus_polltimer = POLLCNT_1SEC;
    BcmHalInterruptEnable(INTERRUPT_ID_EPHY);
    return IRQ_HANDLED;
}

/*
 * bcm63xx_enet_isr: Acknowledge interrupt and check if any packets have
 *                  arrived
 */
static irqreturn_t bcm63xx_enet_isr(int irq, void * dev_id, struct pt_regs * regs)
{
    BcmEnet_devctrl *pDevCtrl = dev_id;

    pDevCtrl->rxDma->intStat = DMA_DONE;  // clr interrupt
    netif_rx_schedule(pDevCtrl->dev);

    return IRQ_HANDLED;
}

/*
 *  bcm63xx_rx: Process all received packets.                  
 */
static uint32 bcm63xx_rx(void *ptr, uint32 budget)
{
    BcmEnet_devctrl *pDevCtrl = ptr;
    struct sk_buff *skb;
    uint16 dmaFlag;
    unsigned char *pBuf;
    int len;
    uint32 rxpkt = 0;

    /* Loop here for each received packet */
    while(pDevCtrl->rxAssignedBds != 0) {
        dmaFlag = pDevCtrl->rxBds[pDevCtrl->rxHeadIndex].status;
        if (dmaFlag & DMA_OWN) {
            rxpkt |= ENET_POLL_DONE;
            break;
        }

        pBuf = (unsigned char *)(phys_to_virt(pDevCtrl->rxBds[pDevCtrl->rxHeadIndex].address));
        len = pDevCtrl->rxBds[pDevCtrl->rxHeadIndex].length;

        if (++pDevCtrl->rxHeadIndex == NR_RX_BDS)
            pDevCtrl->rxHeadIndex = 0;


        local_bh_disable();
        pDevCtrl->rxAssignedBds--;
        __local_bh_enable();

        if (dmaFlag & (EMAC_CRC_ERROR | EMAC_OV | EMAC_NO | EMAC_LG)) {
            pDevCtrl->stats.rx_errors++;
            assign_rx_buffers(pDevCtrl, pBuf - RX_ENET_SKB_RESERVE);
            break;
        }
        else
            if (pDevCtrl->freeSkbList == NULL) {
                pDevCtrl->stats.rx_dropped++;
                assign_rx_buffers(pDevCtrl, pBuf - RX_ENET_SKB_RESERVE);
                break;
            }

        rxpkt++;

        skb = pDevCtrl->freeSkbList;
        pDevCtrl->freeSkbList = pDevCtrl->freeSkbList->retfreeq_context;

        skb_hdrinit(RX_ENET_SKB_RESERVE, SKB_DATA_ALIGN(ENET_MAX_MTU_SIZE), skb,
            pBuf, (void *) bcm63xx_enet_free_skb_or_data, pDevCtrl, 0);
        __skb_trim(skb, len - ETH_CRC_LEN);

#ifdef DUMP_DATA
        printk("bcm63xx_rx :");
        dumpHexData(skb->data, 32);
#endif

        /* Finish setting up the received SKB and send it to the kernel */
        skb->dev = pDevCtrl->dev;

        /* Set the protocol type */
        skb->protocol = eth_type_trans(skb, pDevCtrl->dev);

        pDevCtrl->stats.rx_packets++;
        pDevCtrl->stats.rx_bytes += len;

        /* Notify kernel */
        netif_receive_skb(skb);
        if (--budget <= 0)
            break;
    }

    pDevCtrl->dev->last_rx = jiffies;

    return rxpkt;
}

/*
 * kernel skb free call back
 */
static void bcm63xx_enet_free_skb_or_data(void *param, void *obj, int free_flag)
{
    BcmEnet_devctrl *pDevCtrl = (BcmEnet_devctrl *) param;

    if( free_flag == FREE_SKB ) {
        struct sk_buff *skb = (struct sk_buff *) obj;
        local_bh_disable();
        skb->retfreeq_context = pDevCtrl->freeSkbList;
        pDevCtrl->freeSkbList = skb;
        __local_bh_enable();
    }
    else {
        assign_rx_buffers(pDevCtrl, (unsigned char *) obj);
    }
}

/*
 * Set the hardware MAC address.
 */
static int bcm_set_mac_addr(struct net_device *dev, void *p)
{
    struct sockaddr *addr=p;

    if(netif_running(dev))
        return -EBUSY;

    memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

    write_mac_address(dev);

    return 0;
}

/*
 * write_mac_address: store MAC address into EMAC perfect match registers                   
 */
void write_mac_address(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile uint32 val;

    ASSERT(pDevCtrl != NULL);

    val = pDevCtrl->emac->config;
    if (val & EMAC_ENABLE) {
        /* disable ethernet MAC while updating its registers */
        pDevCtrl->emac->config |= EMAC_DISABLE ;           
        while(pDevCtrl->emac->config & EMAC_DISABLE);     
    }

    /* add our MAC address to the perfect match register */
    perfectmatch_update(pDevCtrl, dev->dev_addr, 1);

    if (val & EMAC_ENABLE) {
        pDevCtrl->emac->config = val;
    }
}


/*
 * assign_rx_buffers: Beginning with the first receive buffer descriptor
 *                  used to receive a packet since this function last
 *                  assigned a buffer, reassign buffers to receive
 *                  buffer descriptors, and mark them as EMPTY; available
 *                  to be used again.
 *
 */
static void assign_rx_buffers(BcmEnet_devctrl *pDevCtrl, unsigned char *data)
{
    local_bh_disable();

    cache_wback_inv((unsigned long)data, RX_BUF_SIZE);

    /* assign packet, prepare descriptor, and advance pointer */
    pDevCtrl->rxBds[pDevCtrl->rxTailIndex].address = (uint32)VIRT_TO_PHY(data + RX_ENET_SKB_RESERVE);
    pDevCtrl->rxBds[pDevCtrl->rxTailIndex].length  = RX_BUF_LEN;

    if (pDevCtrl->unit == 0)
        pDevCtrl->dmaCtrl->flowctl_ch1_alloc = 1;
    else
        pDevCtrl->dmaCtrl->flowctl_ch3_alloc = 1;

    if (pDevCtrl->rxTailIndex == (NR_RX_BDS - 1)) {
        pDevCtrl->rxBds[pDevCtrl->rxTailIndex].status = DMA_OWN | DMA_WRAP;
        pDevCtrl->rxTailIndex = 0;
    } else {
        pDevCtrl->rxBds[pDevCtrl->rxTailIndex].status = DMA_OWN;
        pDevCtrl->rxTailIndex++;
    }

    pDevCtrl->rxAssignedBds++;

    pDevCtrl->rxDma->cfg = DMA_ENABLE;

    __local_bh_enable();
}

/*
 * perfectmatch_write: write an address to the EMAC perfect match registers
 */
static void perfectmatch_write(volatile EmacRegisters *emac, int reg, const uint8 * pAddr, bool bValid)
{
    volatile uint32 *pmDataLo;
    volatile uint32 *pmDataHi;
  
    switch (reg) {
    case 0:
        pmDataLo = &emac->pm0DataLo;
        pmDataHi = &emac->pm0DataHi;
        break;
    case 1:
        pmDataLo = &emac->pm1DataLo;
        pmDataHi = &emac->pm1DataHi;
        break;
    case 2:
        pmDataLo = &emac->pm2DataLo;
        pmDataHi = &emac->pm0DataHi;
        break;
    case 3:
        pmDataLo = &emac->pm3DataLo;
        pmDataHi = &emac->pm3DataHi;
        break;
    default:
        return;
    }
    /* Fill DataHi/Lo */
    if (bValid == 1)
        *pmDataLo = MAKE4((pAddr + 2));
    *pmDataHi = MAKE2(pAddr) | (bValid << 16);
}

/*
 * perfectmatch_update: update an address to the EMAC perfect match registers
 */
static void perfectmatch_update(BcmEnet_devctrl *pDevCtrl, const uint8 * pAddr, bool bValid)
{
    int i;
    unsigned int aged_ref;
    int aged_entry;

    /* check if this address is in used */
    for (i = 0; i < MAX_PMADDR; i++) {
        if (pDevCtrl->pmAddr[i].valid == 1) {
            if (memcmp (pDevCtrl->pmAddr[i].dAddr, pAddr, ETH_ALEN) == 0) {
                if (bValid == 0) {
                    /* clear the valid bit in PM register */
                    perfectmatch_write(pDevCtrl->emac, i, pAddr, bValid);
                    /* clear the valid bit in pDevCtrl->pmAddr */
                    pDevCtrl->pmAddr[i].valid = bValid;
                } else {
                    pDevCtrl->pmAddr[i].ref++;
                }
                return;
            }
        }
    }
    if (bValid == 1) {
        for (i = 0; i < MAX_PMADDR; i++) {
            /* find an available entry for write pm address */
            if (pDevCtrl->pmAddr[i].valid == 0) {
                break;
            }
        }
        if (i == MAX_PMADDR) {
            aged_ref = 0xFFFFFFFF;
            aged_entry = 0;
            /* aged out an entry */
            for (i = 0; i < MAX_PMADDR; i++) {
                if (pDevCtrl->pmAddr[i].ref < aged_ref) {
                    aged_ref = pDevCtrl->pmAddr[i].ref;
                    aged_entry = i;
                }
            }
            i = aged_entry;
        }
        /* find a empty entry and add the address */
        perfectmatch_write(pDevCtrl->emac, i, pAddr, bValid);

        /* update the pDevCtrl->pmAddr */
        pDevCtrl->pmAddr[i].valid = bValid;
        memcpy(pDevCtrl->pmAddr[i].dAddr, pAddr, ETH_ALEN);
        pDevCtrl->pmAddr[i].ref = 1;
    }
}

/*
 * clear_mib: Initialize the Ethernet Switch MIB registers
 */
static void clear_mib(volatile EmacRegisters *emac)
{
    int i;
    volatile uint32 *pt;

    pt = (uint32 *)&emac->tx_mib;
    for( i = 0; i < NumEmacTxMibVars; i++ ) {
        *pt++ = 0;
    }

    pt = (uint32 *)&emac->rx_mib;;
    for( i = 0; i < NumEmacRxMibVars; i++ ) {
        *pt++ = 0;
    }
}

/*
 * init_emac: Initializes the Ethernet Switch control registers
 */
static int init_emac(BcmEnet_devctrl *pDevCtrl)
{
    volatile EmacRegisters *emac;
#if defined(CONFIG_BCM96348)
    uint32 gpiomode;
#endif

    TRACE(("bcm63xxenet: init_emac\n"));

#if defined(CONFIG_BCM96348)
    if (pDevCtrl->EnetInfo.ucPhyType != BP_ENET_INTERNAL_PHY) {
        gpiomode = GPIO->GPIOMode;
        gpiomode |= (GROUP3_EXT_MII|GROUP0_EXT_MII); /* External MII */
        if (pDevCtrl->EnetInfo.usConfigType == BP_ENET_CONFIG_SPI_SSB_1)
            gpiomode |= GROUP1_SPI_MASTER;
        GPIO->GPIOMode = gpiomode;
    }
#endif
#if defined(CONFIG_BCM96338)
    SPI->spiClkCfg = SPI_CLK_0_781MHZ;
#endif

    if (pDevCtrl->unit == 0)
        pDevCtrl->emac = EMAC1;
#if defined(CONFIG_BCM96348)
    else
        pDevCtrl->emac = EMAC2;
#endif

    emac = pDevCtrl->emac;
    /* Initialize the Ethernet Switch MIB registers */
    clear_mib(pDevCtrl->emac);

    /* disable ethernet MAC while updating its registers */
    emac->config = EMAC_DISABLE ;           
    while(emac->config & EMAC_DISABLE);     

    /* issue soft reset, wait for it to complete */
    emac->config = EMAC_SOFT_RESET;
    while (emac->config & EMAC_SOFT_RESET);

    if (mii_init(pDevCtrl->dev))
        return -EFAULT;

    /* Initialize emac registers */
    emac->rxControl = EMAC_FC_EN | EMAC_PROM;   /* allow Promiscuous */

#ifdef MAC_LOOPBACK
    emac->rxControl |= EMAC_LOOPBACK;
#endif
    emac->rxMaxLength = ENET_MAX_MTU_SIZE;
    emac->txMaxLength = ENET_MAX_MTU_SIZE;

    /* tx threshold = abs(63-(0.63*EMAC_DMA_MAX_BURST_LENGTH)) */
    emac->txThreshold = EMAC_TX_WATERMARK;
    emac->mibControl  = EMAC_NO_CLEAR;
    emac->intMask = 0;              /* mask all EMAC interrupts*/

    return 0;
}

/*
 * init_dma: Initialize DMA control register
 */
static void init_dma(BcmEnet_devctrl *pDevCtrl)
{
    DmaStateRam *StateRam;

    TRACE(("bcm63xxenet: init_dma\n"));
    /*
     * clear State RAM
     */
    if (pDevCtrl->unit == 0)
        StateRam = (DmaStateRam *)&pDevCtrl->dmaCtrl->stram.s[EMAC1_RX_CHAN];
    else
        StateRam = (DmaStateRam *)&pDevCtrl->dmaCtrl->stram.s[EMAC2_RX_CHAN];
    memset(StateRam, 0, sizeof(DmaStateRam) * NUM_CHANS_PER_EMAC);

    /*
     * initialize IUDMA controller register
     */
    if (pDevCtrl->unit == 0) {
        pDevCtrl->dmaCtrl->controller_cfg |= DMA_FLOWC_CH1_EN;
        pDevCtrl->dmaCtrl->flowctl_ch1_thresh_lo = DMA_FC_THRESH_LO;
        pDevCtrl->dmaCtrl->flowctl_ch1_thresh_hi = DMA_FC_THRESH_HI;
    } else {
        pDevCtrl->dmaCtrl->controller_cfg |= DMA_FLOWC_CH3_EN;
        pDevCtrl->dmaCtrl->flowctl_ch3_thresh_lo = DMA_FC_THRESH_LO;
        pDevCtrl->dmaCtrl->flowctl_ch3_thresh_hi = DMA_FC_THRESH_HI;
    }

    /* setup transmit dma register */
    pDevCtrl->txDma->cfg = 0;       /* initialize first (will enable later) */
    pDevCtrl->txDma->maxBurst = DMA_MAX_BURST_LENGTH;
    pDevCtrl->txDma->intMask = 0;   /* mask all ints */
    if (pDevCtrl->unit == 0)
        pDevCtrl->dmaCtrl->stram.s[EMAC1_TX_CHAN].baseDescPtr = 
            (uint32)VIRT_TO_PHY((uint32 *)pDevCtrl->txBds);
    else
        pDevCtrl->dmaCtrl->stram.s[EMAC2_TX_CHAN].baseDescPtr = 
            (uint32)VIRT_TO_PHY((uint32 *)pDevCtrl->txBds);

    /* setup receive dma register */
    pDevCtrl->rxDma->cfg = 0;  // initialize first (will enable later)
    pDevCtrl->rxDma->maxBurst = DMA_MAX_BURST_LENGTH;
    pDevCtrl->rxDma->intMask = 0;   /* mask all ints */
    /* clr any pending interrupts on channel */
    pDevCtrl->rxDma->intStat = DMA_DONE;
    /* set to interrupt on packet complete and no descriptor available */
    pDevCtrl->rxDma->intMask = DMA_DONE;
    if (pDevCtrl->unit == 0)
        pDevCtrl->dmaCtrl->stram.s[EMAC1_RX_CHAN].baseDescPtr = 
            (uint32)VIRT_TO_PHY((uint32 *)pDevCtrl->rxBds);
    else
        pDevCtrl->dmaCtrl->stram.s[EMAC2_RX_CHAN].baseDescPtr = 
            (uint32)VIRT_TO_PHY((uint32 *)pDevCtrl->rxBds);
}

/*
 *  init_buffers: initialize driver's pools of receive buffers
 *  and tranmit headers
 */
static int init_buffers(BcmEnet_devctrl *pDevCtrl)
{
    unsigned char *data;
    int i;

    TRACE(("bcm63xxenet: init_buffers\n"));

    /* allocate recieve buffer pool */
    for (i = 0; i < NR_RX_BDS; i++) {
        data = kmalloc(RX_BUF_SIZE, GFP_KERNEL);
        if (data == NULL) {
            printk(KERN_NOTICE CARDNAME": Low memory.\n");
            break;
        }
        pDevCtrl->buf_pool[i] = data;
        assign_rx_buffers(pDevCtrl, data);
    }

    if (i < NR_RX_BDS) {
        /* release allocated receive buffer memory */
        for (i = 0; i < NR_RX_BDS; i++) {
            if (pDevCtrl->buf_pool[i]) {
                kfree(pDevCtrl->buf_pool[i]);
                pDevCtrl->buf_pool[i] = NULL;
            }
        }
        return -ENOMEM;
    }
    return 0;
}

/*
 * bcm63xx_init_dev: initialize Ethernet MACs,
 * allocate Tx/Rx buffer descriptors pool, Tx header pool.
 */
static int bcm63xx_init_dev(BcmEnet_devctrl *pDevCtrl)
{
    void *p;
    int i;
    unsigned char *skb;

    TRACE(("bcm63xxenet: bcm63xx_init_dev\n"));

    /* make sure emac clock is on */
    PERF->blkEnables |= EMAC_CLK_EN;

    /* init rx/tx dma channels */
    pDevCtrl->dmaCtrl = (DmaRegs *)(EMAC_DMA_BASE);
    if (pDevCtrl->unit == 0) {
        pDevCtrl->rxDma = &pDevCtrl->dmaCtrl->chcfg[EMAC1_RX_CHAN];
        pDevCtrl->txDma = &pDevCtrl->dmaCtrl->chcfg[EMAC1_TX_CHAN];
        pDevCtrl->rxIrq = INTERRUPT_ID_EMAC1_RX_DMA;
    } else {
#if defined(CONFIG_BCM96348)
        pDevCtrl->rxDma = &pDevCtrl->dmaCtrl->chcfg[EMAC2_RX_CHAN];
        pDevCtrl->txDma = &pDevCtrl->dmaCtrl->chcfg[EMAC2_TX_CHAN];
        pDevCtrl->rxIrq = INTERRUPT_ID_EMAC2_RX_DMA;
#endif
    }
    BcmHalInterruptDisable(pDevCtrl->rxIrq);

    /* register the interrupt service handler */
    BcmHalMapInterrupt(bcm63xx_enet_isr, (unsigned int)pDevCtrl, pDevCtrl->rxIrq);

    if (pDevCtrl->EnetInfo.ucPhyType == BP_ENET_INTERNAL_PHY) {
        BcmHalInterruptDisable(INTERRUPT_ID_EPHY);
        BcmHalMapInterrupt(bcm63xx_enet_phy_isr, (unsigned int)pDevCtrl,
            INTERRUPT_ID_EPHY);
    }

    /* setup buffer/pointer relationships here */
    /* allocate and assign tx buffer descriptors */
    if (!(p = kmalloc(NR_TX_BDS*sizeof(DmaDesc), GFP_KERNEL))) {
        return -ENOMEM;
    }
    memset(p, 0, NR_TX_BDS*sizeof(DmaDesc));
    cache_wback_inv((unsigned long)p, NR_TX_BDS*sizeof(DmaDesc));
    pDevCtrl->txBds = (DmaDesc *)CACHE_TO_NONCACHE(p);   /* tx bd ring */

    /* alloc space for the rx buffer descriptors */
    if (!(p = kmalloc(NR_RX_BDS*sizeof(DmaDesc), GFP_KERNEL))) {
        kfree((void *)NONCACHE_TO_CACHE(pDevCtrl->txBds));
        return -ENOMEM;
    }
    memset(p, 0, NR_RX_BDS*sizeof(DmaDesc));
    cache_wback_inv((unsigned long)p, NR_RX_BDS*sizeof(DmaDesc));
    pDevCtrl->rxBds = (DmaDesc *)CACHE_TO_NONCACHE(p);   /* rx bd ring */

    pDevCtrl->rxAssignedBds = 0;
    pDevCtrl->rxHeadIndex = pDevCtrl->rxTailIndex = 0;

    pDevCtrl->txFreeBds = NR_TX_BDS;
    pDevCtrl->txHeadIndex = pDevCtrl->txTailIndex = 0;

    /* initialize the receive buffers and transmit headers */
    if (init_buffers(pDevCtrl)) {
        kfree((void *)NONCACHE_TO_CACHE(pDevCtrl->txBds));
        kfree((void *)NONCACHE_TO_CACHE(pDevCtrl->rxBds));
        return -ENOMEM;
    }

    /* Chain socket buffers. */
    for(i = 0, skb = (unsigned char *)
        (((unsigned long) pDevCtrl->skbs + 0x0f) & ~0x0f); i < NR_RX_BDS;
        i++, skb += SKB_ALIGNED_SIZE)
    {
        ((struct sk_buff *) skb)->retfreeq_context = pDevCtrl->freeSkbList;
        pDevCtrl->freeSkbList = (struct sk_buff *) skb;
    }

    /* init dma registers */
    init_dma(pDevCtrl);

    /* init switch control registers */
    if (init_emac(pDevCtrl))
        return -EFAULT;

    /* if we reach this point, we've init'ed successfully */
    return 0;
}

/* Uninitialize tx/rx buffer descriptor pools */
static int bcm63xx_uninit_dev(BcmEnet_devctrl *pDevCtrl)
{
    char proc_name[32];
    int i;

    if (pDevCtrl) {
        del_timer_sync(&pDevCtrl->timer);

        /* disable DMA */
        pDevCtrl->txDma->cfg = 0;
        pDevCtrl->rxDma->cfg = 0;
        if (pDevCtrl->unit == 0)
            pDevCtrl->dmaCtrl->flowctl_ch1_alloc = (DMA_BUF_ALLOC_FORCE | 0);
        else
            pDevCtrl->dmaCtrl->flowctl_ch3_alloc = (DMA_BUF_ALLOC_FORCE | 0);

       /* free the irq */
        if (pDevCtrl->rxIrq) {
            BcmHalInterruptDisable(pDevCtrl->rxIrq);
            free_irq(pDevCtrl->rxIrq, pDevCtrl);
        }
        if (pDevCtrl->EnetInfo.ucPhyType == BP_ENET_INTERNAL_PHY) {
            BcmHalInterruptDisable(INTERRUPT_ID_EPHY);
            free_irq(INTERRUPT_ID_EPHY, pDevCtrl);
        }
        /* free the skbs */
        while (pDevCtrl->txFreeBds < NR_TX_BDS)  {
            pDevCtrl->txFreeBds++;
            dev_kfree_skb_any (pDevCtrl->txSkb[pDevCtrl->txHeadIndex++]);
            if (pDevCtrl->txHeadIndex == NR_TX_BDS)
                pDevCtrl->txHeadIndex = 0;
        }
        /* release allocated receive buffer memory */
        for (i = 0; i < NR_RX_BDS; i++) {
            if (pDevCtrl->buf_pool[i]) {
                kfree(pDevCtrl->buf_pool[i]);
                pDevCtrl->buf_pool[i] = NULL;
            }
        }
        /* free the transmit buffer descriptor */
        if (pDevCtrl->txBds)
            kfree((void *)NONCACHE_TO_CACHE(pDevCtrl->txBds));
        /* free the receive buffer descriptor */
        if (pDevCtrl->rxBds)
            kfree((void *)NONCACHE_TO_CACHE(pDevCtrl->rxBds));
#ifdef USE_PROC
        sprintf(proc_name, PROC_ENTRY_NAME, pDevCtrl->unit);
        remove_proc_entry(proc_name, NULL);
#endif
        if (pDevCtrl->dev) {
            if (pDevCtrl->dev->reg_state != NETREG_UNINITIALIZED)
                unregister_netdev(pDevCtrl->dev);
            free_netdev(pDevCtrl->dev);
        }
    }
   
    return 0;
}

#ifdef USE_PROC
/*
 * bcm63xx_enet_dump - display EMAC information
 */
static int bcm63xx_enet_dump(BcmEnet_devctrl *pDevCtrl, char *buf, int reqinfo)
{
    volatile EmacRegisters *emac;
    DmaStateRam *StateRam;
    int val;
    int i;
    int len = 0;
    int bufcnt;

    switch (reqinfo) {
    case TX_BD_RING1: /* tx DMA BD descriptor ring */
        if (pDevCtrl->unit == 0)
            len += sprintf(&buf[len], "\nEMAC1\n");
        else
            len += sprintf(&buf[len], "\nEMAC2\n");
        len += sprintf(&buf[len], "\ntx buffer descriptor ring status.\n");
        len += sprintf(&buf[len], "BD\tlocation\tlength\tstatus\n");
        for (i = 0; i < NR_TX_BDS/2; i++) {
            len += sprintf(&buf[len], "%03d\t%08x\t%04d\t%04x\n",
                   i,(unsigned int)&pDevCtrl->txBds[i],
                   pDevCtrl->txBds[i].length,
                   pDevCtrl->txBds[i].status);
        }
        break;

    case TX_BD_RING2: /* tx DMA BD descriptor ring */
        for (i = NR_TX_BDS/2; i < NR_TX_BDS; i++) {
            len += sprintf(&buf[len], "%03d\t%08x\t%04d\t%04x\n",
                   i,(unsigned int)&pDevCtrl->txBds[i],
                   pDevCtrl->txBds[i].length,
                   pDevCtrl->txBds[i].status);
        }
        break;

    case RX_BD_RING: /* rx DMA BD descriptor ring */
        len += sprintf(&buf[len], "\nrx buffer descriptor ring status.\n");
        len += sprintf(&buf[len], "BD\tlocation\tlength\tstatus\n");
        for (i = 0; i < NR_RX_BDS; i++) {
            len += sprintf(&buf[len], "%03d\t%08x\t%04d\t%04x\n",
                   i,(int)&pDevCtrl->rxBds[i],
                   pDevCtrl->rxBds[i].length,
                   pDevCtrl->rxBds[i].status);
        }
        break;

    case DMA_DESC_STATUS:  /* DMA descriptors pointer and status */
        len += sprintf(&buf[len], "\nrx pointers:\n");
        len += sprintf(&buf[len], "DmaDesc *rxBds:\t\t\t%08x\n", (int) pDevCtrl->rxBds);
        len += sprintf(&buf[len], "DmaDesc *rxHeadIndex:\t\t%d\n", pDevCtrl->rxHeadIndex);
        len += sprintf(&buf[len], "DmaDesc *rxTailIndex:\t\t%d\n", pDevCtrl->rxTailIndex);
        len += sprintf(&buf[len], "rxAssignedBds (RxBDs %d):\t\t%d\n", NR_RX_BDS, pDevCtrl->rxAssignedBds);

        len += sprintf(&buf[len], "\ntx pointers:\n");
        len += sprintf(&buf[len], "DmaDesc *txBds:\t\t\t%08x\n", (int)pDevCtrl->txBds);
        len += sprintf(&buf[len], "DmaDesc *txHeadIndex:\t\t%d\n", pDevCtrl->txHeadIndex);
        len += sprintf(&buf[len], "DmaDesc *txTailIndex:\t\t%d\n", pDevCtrl->txTailIndex);
        len += sprintf(&buf[len], "txFreeBds (TxBDs %d):\t\t%d\n", NR_TX_BDS, pDevCtrl->txFreeBds);

        len += sprintf(&buf[len], "\ntx DMA Channel Config\t\t%08lx\n", pDevCtrl->txDma->cfg);
        len += sprintf(&buf[len], "tx DMA Intr Control/Status\t%08lx\n", pDevCtrl->txDma->intStat);
        len += sprintf(&buf[len], "tx DMA Intr Mask\t\t%08lx\n", pDevCtrl->txDma->intMask);

        len += sprintf(&buf[len], "\nrx DMA Channel Config\t\t%08lx\n", pDevCtrl->rxDma->cfg);
        len += sprintf(&buf[len], "rx DMA Intr Control/Status\t%08lx\n", pDevCtrl->rxDma->intStat);
        len += sprintf(&buf[len], "rx DMA Intr Mask\t\t%08lx\n", pDevCtrl->rxDma->intMask);
        if (pDevCtrl->unit == 0) {
            len += sprintf(&buf[len], "rx DMA Controller Configuration\t%08lx\n", pDevCtrl->dmaCtrl->controller_cfg);
            len += sprintf(&buf[len], "rx DMA Low Threshold\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch1_thresh_lo);
            len += sprintf(&buf[len], "rx DMA High Threshold\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch1_thresh_hi);
            len += sprintf(&buf[len], "rx DMA Buffer Alloc\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch1_alloc);
        } else {
            len += sprintf(&buf[len], "rx DMA Controller Configuration\t%08lx\n", pDevCtrl->dmaCtrl->controller_cfg);
            len += sprintf(&buf[len], "rx DMA Low Threshold\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch3_thresh_lo);
            len += sprintf(&buf[len], "rx DMA High Threshold\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch3_thresh_hi);
            len += sprintf(&buf[len], "rx DMA Buffer Alloc\t\t%ld\n", pDevCtrl->dmaCtrl->flowctl_ch3_alloc);
        }
        break;

    case BUF_POOL: /* buffer usage */
        bufcnt = 0;
        len += sprintf(&buf[len], "\nbuf\taddress\n");
        for (i = 0; i < NR_RX_BDS; i++) {
            len += sprintf(&buf[len], "%d\t%08x\n", i, (int)pDevCtrl->buf_pool[i]);
        }
        break;

    case EMAC_REGS: /* EMAC registers */
        emac = pDevCtrl->emac;
        len += sprintf(&buf[len], "\nEMAC registers\n");
        len += sprintf(&buf[len], "rx config register\t0x%08lx\n", emac->rxControl);
        len += sprintf(&buf[len], "rx max length register\t0x%08lx\n", emac->rxMaxLength);
        len += sprintf(&buf[len], "tx max length register\t0x%08lx\n", emac->txMaxLength);
        len += sprintf(&buf[len], "interrupt mask register\t0x%08lx\n", emac->intMask);
        len += sprintf(&buf[len], "interrupt register\t0x%08lx\n", emac->intStatus);
        len += sprintf(&buf[len], "control register\t0x%08lx\n", emac->config);
        len += sprintf(&buf[len], "tx control register\t0x%08lx\n", emac->txControl);
        len += sprintf(&buf[len], "tx watermark register\t0x%08lx\n", emac->txThreshold);
        break;

    case MISC_REGS: /* misc registers */
        if (pDevCtrl->unit == 0)
            StateRam = (DmaStateRam *)&pDevCtrl->dmaCtrl->stram.s[EMAC1_RX_CHAN];
        else
            StateRam = (DmaStateRam *)&pDevCtrl->dmaCtrl->stram.s[EMAC2_RX_CHAN];
        len += sprintf(&buf[len], "ring start address\t0x%08lx\n", StateRam->baseDescPtr);
        len += sprintf(&buf[len], "state/bytes offset\t0x%08lx\n", StateRam->state_data);
        len += sprintf(&buf[len], "BD status and len\t0x%08lx\n", StateRam->desc_len_status);
        len += sprintf(&buf[len], "BD current processing\t0x%08lx\n", StateRam->desc_base_bufptr);
        val = (read_c0_cause() & CAUSEF_IP);
        len += sprintf(&buf[len], "CP0 cause\t\t0x%08x\n", val);
        val = read_c0_status();
        len += sprintf(&buf[len], "CP0 status\t\t0x%08x\n", val);
        len += sprintf(&buf[len], "PERF->IrqMask\t\t0x%08lx\n", PERF->IrqMask);
        len += sprintf(&buf[len], "PERF->IrqStatus\t\t0x%08lx\n", PERF->IrqStatus);
        break;

    default:
        break;
    }

    return len;
}

static int dev_proc_engine(void *data,loff_t pos, char *buf)
{
    BcmEnet_devctrl *pDevCtrl = (BcmEnet_devctrl *)data;
    int len = 0;

    if (pDevCtrl == NULL)
        return len;

    switch((int)pos) {
    case TX_BD_RING1: /* tx DMA BD descriptor ring */
        len += bcm63xx_enet_dump(pDevCtrl, buf, TX_BD_RING1);
        break;
    case TX_BD_RING2: /* tx DMA BD descriptor ring */
        len += bcm63xx_enet_dump(pDevCtrl, buf, TX_BD_RING2);
        break;
    case RX_BD_RING: /* rx DMA BD descriptor ring */
        len += bcm63xx_enet_dump(pDevCtrl, buf, RX_BD_RING);
        break;
    case DMA_DESC_STATUS: /* DMA descriptors pointer and status */
        len += bcm63xx_enet_dump(pDevCtrl, buf, DMA_DESC_STATUS);
        break;
    case BUF_POOL: /* buffer usage */
        len += bcm63xx_enet_dump(pDevCtrl, buf, BUF_POOL);
        break;
    case EMAC_REGS: /* EMAC registers */
        len += bcm63xx_enet_dump(pDevCtrl, buf, EMAC_REGS);
        break;
    case MISC_REGS: /* misc registers */
        len += bcm63xx_enet_dump(pDevCtrl, buf, MISC_REGS);
        break;
    default:
        break;
    }
    return len;
}

/*
 *  read proc interface
 */
static ssize_t eth_proc_read(struct file *file, char *buf, size_t count,
        loff_t *pos)
{
    const struct inode *inode = file->f_dentry->d_inode;
    const struct proc_dir_entry *dp = PDE(inode);
    char *page;
    int len = 0, x, left;

    page = kmalloc(PAGE_SIZE, GFP_KERNEL);
    if (!page)
        return -ENOMEM;
    left = PAGE_SIZE - 256;
    if (count < left)
        left = count;

    for (;;) {
        x = dev_proc_engine(dp->data, *pos, &page[len]);
        if (x == 0)
            break;
        if ((x + len) > left)
            x = -EINVAL;
        if (x < 0) {
            break;
        }
        len += x;
        left -= x;
        (*pos)++;
        if (left < 256)
            break;
    }
    if (len > 0 && copy_to_user(buf, page, len))
        len = -EFAULT;
    kfree(page);
    return len;
}

/*
 * /proc/driver/eth0info
 */
static struct file_operations eth_proc_operations = {
        read: eth_proc_read, /* read_proc */
};

#endif

/*
 *      bcm63xx_enet_probe: - Probe Ethernet switch and allocate device
 */
int __init bcm63xx_enet_probe(void)
{
    static int probed = 0;
    struct net_device *dev = NULL;
    BcmEnet_devctrl *pDevCtrl = NULL;
    unsigned int chipid;
    unsigned int chiprev;
    unsigned char macAddr[ETH_ALEN];
    ETHERNET_MAC_INFO EnetInfo[MAX_EMACS];
    int status;
    int unit;
    int i;
#ifdef USE_PROC
    struct proc_dir_entry *p;
    char proc_name[32];
#endif

    TRACE(("bcm63xxenet: bcm63xx_enet_probe\n"));

    if (probed == 0) {
        chipid  = (PERF->RevID & 0xFFFF0000) >> 16;
        chiprev = (PERF->RevID & 0xFF);
        /* figure out which chip we're running on */
        switch (chipid) {
            case 0x6338:
                break;
            case 0x6348:
                break;
            default:
                printk(KERN_DEBUG CARDNAME" not found\n");
                return -ENODEV;
        }
        if( BpGetEthernetMacInfo( &EnetInfo[0], MAX_EMACS ) != BP_SUCCESS ) {
            printk(KERN_DEBUG CARDNAME" board id not set\n");
            return -ENODEV;
        }
        probed++;
    } else {
        /* device has already been initialized */
        return -ENXIO;
    }

    for ( unit = 0; unit < MAX_EMACS; unit++ ) {
        switch(EnetInfo[unit].ucPhyType) {
            case BP_ENET_INTERNAL_PHY:
            case BP_ENET_EXTERNAL_PHY:
            case BP_ENET_EXTERNAL_SWITCH:
                break;

            case BP_ENET_NO_PHY:
            default:
                continue;
        }
        dev = alloc_etherdev(sizeof(*pDevCtrl));
        if (dev == NULL) {
            printk(KERN_ERR CARDNAME": Unable to allocate "
                    "net_device structure!\n");
            continue;
        }
        pDevCtrl = netdev_priv(dev);
        pDevCtrl->dev = dev;
        pDevCtrl->unit = unit;
        pDevCtrl->chipId  = chipid;
        pDevCtrl->chipRev = chiprev;
        pDevCtrl->linkstatus_phyport = -1;
        memcpy(&(pDevCtrl->EnetInfo), &EnetInfo[unit], sizeof(ETHERNET_MAC_INFO));
        /* print the ChipID and module version info */
        printk("Broadcom BCM%X%X Ethernet Network Device ", chipid, chiprev);
        printk(VER_STR);
        printk("\n");

        if ((status = bcm63xx_init_dev(pDevCtrl))) {
            printk((KERN_ERR CARDNAME ": device initialization error!\n"));
            bcm63xx_uninit_dev(pDevCtrl);
            continue;
        }
#ifdef USE_PROC
        /* create a /proc entry for display driver runtime information */
        sprintf(proc_name, PROC_ENTRY_NAME, pDevCtrl->unit);
        if ((p = create_proc_entry(proc_name, 0, NULL)) == NULL) {
            printk((KERN_ERR CARDNAME ": unable to create proc entry!\n"));
            bcm63xx_uninit_dev(pDevCtrl);
            continue;
        }
        p->proc_fops = &eth_proc_operations;
        p->data = pDevCtrl;
#endif

        dev_alloc_name(dev, dev->name);
        SET_MODULE_OWNER(dev);

        pDevCtrl->next_dev = eth_root_dev;
        eth_root_dev = dev;
        /* Setup the timer */
        init_timer(&pDevCtrl->timer);
        pDevCtrl->timer.data = (unsigned long)pDevCtrl;
        pDevCtrl->timer.function = bcm63xx_enet_poll_timer;

        if (pDevCtrl->unit == 0)
            dev->base_addr      = EMAC1_BASE;
#if defined(CONFIG_BCM96348)
        else
            dev->base_addr      = EMAC2_BASE;
#endif
        dev->irq                = pDevCtrl->rxIrq;
        dev->open               = bcm63xx_enet_open;
        dev->stop               = bcm63xx_enet_close;
        dev->hard_start_xmit    = bcm63xx_enet_xmit;
        dev->tx_timeout         = bcm63xx_enet_timeout;
        dev->watchdog_timeo     = 2*HZ;
        dev->get_stats          = bcm63xx_enet_query;
        dev->set_mac_address    = bcm_set_mac_addr;
        dev->set_multicast_list = bcm_set_multicast_list;
        dev->do_ioctl           = &bcm63xx_enet_ioctl;
        dev->poll               = bcm63xx_enet_poll;
        dev->weight             = 64;

        status = register_netdev(dev);
        if (status == 0) {
            probed++;
 
            /* Read MAC address */
            macAddr[0] = 0xff;
            kerSysGetMacAddress( macAddr, dev->ifindex );
            if( (macAddr[0] & ETH_MULTICAST_BIT) == ETH_MULTICAST_BIT ) {
                memcpy( macAddr, "\x00\x10\x18\x63\x00\x00", ETH_ALEN );
                printk((KERN_CRIT "%s: MAC address has not been initialized in NVRAM.\n"),dev->name);
            }

            /* fill in the MAC address */
            for (i = 0; i < ETH_ALEN; i++) {
                dev->dev_addr[i] = macAddr[i];
            }
            write_mac_address(dev);

            /* print the Ethenet address */
            printk("%s: MAC Address: ", dev->name);
            for (i = 0; i < ETH_ALEN-1; i++) {
                printk("%2.2X:", dev->dev_addr[i]);
            }
            printk("%2.2X\n", dev->dev_addr[i]);

            /* Start the poll timer. */
            pDevCtrl->timer.expires = jiffies + POLLTIME_100MS;
            add_timer(&pDevCtrl->timer);

            /* Enable the internal BCM63xx PHY interrupt. */
            if (pDevCtrl->EnetInfo.ucPhyType == BP_ENET_INTERNAL_PHY) {
                mii_enablephyinterrupt(dev, pDevCtrl->EnetInfo.ucPhyAddress);
                BcmHalInterruptEnable(INTERRUPT_ID_EPHY);
            }

            continue;
        } else {
            bcm63xx_uninit_dev(pDevCtrl);
            printk(KERN_ERR CARDNAME "bcm63xx_enet_probe failed, returns %d\n", status);
            return status;
        }
    }
    return 0;
}

/* get ethernet port's status; return nonzero for Link up, 0 for Link down */
static int bcmIsEnetUp(struct net_device *dev)
{
    static int sem = 0;
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    int linkState = pDevCtrl->linkState;

    if( sem == 0 )
    {
        sem = 1;

        if( pDevCtrl->EnetInfo.ucPhyType == BP_ENET_EXTERNAL_SWITCH )
        {
            /* Call non-blocking functions to read the link state. */
            if( pDevCtrl->linkstatus_phyport == -1 )
            {
                /* Start reading the link state of each switch port. */
                pDevCtrl->linkstatus_phyport = 0;
                mii_linkstatus_start(dev, pDevCtrl->EnetInfo.ucPhyAddress |
                    pDevCtrl->linkstatus_phyport);
            }
            else
            {
                int ls = 0;

                /* Check if the link state is done being read for a particular
                 * switch port.
                 */
                if( mii_linkstatus_check(dev, &ls) )
                {
                    /* The link state for one switch port has been read.  Save
                     * it in a temporary holder.
                     */
                    pDevCtrl->linkstatus_holder |=
                        ls << pDevCtrl->linkstatus_phyport++;

                    if( pDevCtrl->linkstatus_phyport ==
                        pDevCtrl->EnetInfo.numSwitchPorts )
                    {
                        /* The link state for all switch ports has been read.
                         * Return the current link state.
                         */
                        linkState = (pDevCtrl->linkstatus_holder) ? 1 : 0;
                        pDevCtrl->linkstatus_holder = 0;
                        pDevCtrl->linkstatus_polltimer = 0;
                        pDevCtrl->linkstatus_phyport = -1;
                    }
                    else
                    {
                        /* Start to read the link state for the next switch
                         * port.
                         */
                        mii_linkstatus_start(dev,
                            pDevCtrl->EnetInfo.ucPhyAddress |
                            pDevCtrl->linkstatus_phyport);
                    }
                }
                else
                    if( pDevCtrl->linkstatus_polltimer > (3 * POLLCNT_1SEC) )
                    {
                        /* Timeout reading MII status. Reset link state check. */
                        pDevCtrl->linkstatus_holder = 0;
                        pDevCtrl->linkstatus_polltimer = 0;
                        pDevCtrl->linkstatus_phyport = -1;
                    }
            }

        }
        else
        {
            /* The link state for the internal PHY is only read after an
             * interrupt indicates that it has changed.  Therefore, it is OK
             * to call the blocking link state read function.
             */
            linkState = mii_linkstatus(dev, pDevCtrl->EnetInfo.ucPhyAddress);
            pDevCtrl->linkstatus_polltimer = POLLCNT_FOREVER;
        }

        sem = 0;
    }

    return linkState;
}

static void __exit bcmenet_module_cleanup(void)
{
    BcmEnet_devctrl *pDevCtrl;
    struct net_device *next_dev;

    TRACE(("bcm63xxenet: bcmenet_module_cleanup\n"));

    /* No need to check MOD_IN_USE, as sys_delete_module() checks. */
    while (eth_root_dev) {
        next_dev = NULL;
        pDevCtrl = (BcmEnet_devctrl *)netdev_priv(eth_root_dev);
        if (pDevCtrl) {
            next_dev = pDevCtrl->next_dev;
            bcm63xx_uninit_dev(pDevCtrl);
        }
        eth_root_dev = next_dev;
    }
}

static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    u32 ethcmd;
    
    ASSERT(pDevCtrl != NULL);   
        
    if (copy_from_user(&ethcmd, useraddr, sizeof(ethcmd)))
        return -EFAULT;
    
    switch (ethcmd) {
    /* get driver-specific version/etc. info */
    case ETHTOOL_GDRVINFO: {
        struct ethtool_drvinfo info = {ETHTOOL_GDRVINFO};

        strncpy(info.driver, CARDNAME, sizeof(info.driver)-1);
        strncpy(info.version, VERSION, sizeof(info.version)-1);
        if (copy_to_user(useraddr, &info, sizeof(info)))
            return -EFAULT;
        return 0;
        }
    default:
        break;
    }
    
    return -EOPNOTSUPP;    
}   

static int bcm63xx_enet_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    int *data=(int*)rq->ifr_data;
    struct mii_ioctl_data *mii;
    int val = 0;

    /* we can add sub-command in ifr_data if we need to in the future */
    switch (cmd)
    {
    case SIOCGMIIPHY:       /* Get address of MII PHY in use. */
        mii = (struct mii_ioctl_data *)&rq->ifr_data;
        mii->phy_id = pDevCtrl->EnetInfo.ucPhyAddress;

    case SIOCGMIIREG:       /* Read MII PHY register. */
        mii = (struct mii_ioctl_data *)&rq->ifr_data;
        mii->val_out = mii_read(dev, mii->phy_id & 0x1f, mii->reg_num & 0x1f);
        break;

    case SIOCSMIIREG:       /* Write MII PHY register. */
        mii = (struct mii_ioctl_data *)&rq->ifr_data;
        mii_write(dev, mii->phy_id & 0x1f, mii->reg_num & 0x1f, mii->val_in);
        break;

    case SIOCGLINKSTATE:
        val = pDevCtrl->linkState;
        if (copy_to_user((void*)data, (void*)&val, sizeof(int))) {
            return -EFAULT;
        }
        val = 0;
        break;

    case SIOCSCLEARMIBCNTR:
        ASSERT(pDevCtrl != NULL);
        memset(&pDevCtrl->stats, 0, sizeof(struct net_device_stats));
        break;

    case SIOCMIBINFO:
        if (copy_to_user((void*)data, (void*)&pDevCtrl->MibInfo,
            sizeof(pDevCtrl->MibInfo))) {
            return -EFAULT;
        }
        break;

    case SIOCETHTOOL:
        return netdev_ethtool_ioctl(dev, (void *) rq->ifr_data);

    case SIOCGENABLEVLAN:
        pDevCtrl->vid = 2;
        val = ethsw_config_vlan(dev, (int)VLAN_ENABLE, pDevCtrl->vid);
        break;

    case SIOCGDISABLEVLAN:
        pDevCtrl->vid = 0;
        val = ethsw_config_vlan(dev, VLAN_DISABLE, 0);
        break;

    case SIOCGQUERYNUMVLANPORTS:
        switch(pDevCtrl->ethSwitch.type) {
        case ESW_TYPE_BCM5325M:
            val = 1;
            break;
        case ESW_TYPE_BCM5325E:
        case ESW_TYPE_BCM5325F:
        default:
            val = pDevCtrl->EnetInfo.numSwitchPorts;
            break;
        }
        if (copy_to_user((void*)data, (void*)&val, sizeof(int))) {
            return -EFAULT;
        }
        val = 0;
        break;

    case SIOCGSWITCHTYPE:
        if (pDevCtrl->EnetInfo.ucPhyType == BP_ENET_EXTERNAL_SWITCH) {
            if (copy_to_user((void*)data, (void*)&pDevCtrl->ethSwitch.type, sizeof(int))) {
                return -EFAULT;
            }
        } else
            return -EFAULT;
        break;

    case SIOCGQUERYNUMPORTS:
        val = pDevCtrl->EnetInfo.numSwitchPorts;
        if (copy_to_user((void*)data, (void*)&val, sizeof(int))) {
            return -EFAULT;
        }
        val = 0;
        break;
    }
    return val;
}

static int __init bcmenet_module_init(void)
{
    int status;

    TRACE(("bcm63xxenet: bcmenet_module_init\n"));

    status = bcm63xx_enet_probe();

    return status;
}

module_init(bcmenet_module_init);
module_exit(bcmenet_module_cleanup);
MODULE_LICENSE("Proprietary");

